# Prot386 Text output and input functions
# (C) 2006-2007 Peter Ambroz
# This code is free software and is distributed under the terms of GNU GPL

	.text
	.globl clrscr
	.globl gotoxy
	.globl hide_cursor, show_cursor
	.globl highlight
	.globl write, put_char
	.globl window, get_window
	.globl save_screen, load_screen
	.globl inttostr, hextostr, strtoint, strtohex
	.globl pad_right
	.globl readkey, read_enter, read_char, read_string
	.globl textattr, clearchar, pos_x, pos_y, win_x1, win_y1, win_x2, win_y2
	.globl use_eol

textattr: .long 7
clearchar: .long 32
pos_x:	.long 0
pos_y:	.long 0
win_x1:	.long 0
win_y1:	.long 0
win_x2:	.long 79
win_y2:	.long 24
use_eol: .long 0

# void clrscr(void)
# clear the text screen with clearchar and reset cursor to the upper left corner
clrscr:
	push %es
	push %gs
	pop %es
	push %edi
	movw clearchar, %ax	# char
	movb textattr, %ah	# color
	movl $25*80, %ecx	# text screen size
	xorl %edi, %edi		# ES:EDI = starting offset of text mode video-ram
	rep stosw		# store color+char information to every position
	# next we will reset cursor positon
	xorw %ax, %ax
	movb $0x0E, %al
	movw $0x3D4, %dx
	outw %ax, %dx		# output value 0x00 to CRT register 0x0E (16 bit port 0x3D4)
	movb $0x0F, %al
	outw %ax, %dx		# output value 0x00 to CRT register 0x0F
	# adjust local cursor position variables
	movl $0, pos_x
	movl $0, pos_y
	pop %edi
	pop %es
	ret

# void gotoxy(int x, int y)
# Move text cursor to new position
gotoxy:
	push %ebp
	movl %esp, %ebp
	
	push %ebx
	movl 8(%ebp), %ebx
	movl 12(%ebp), %edx
	
	addl win_x1, %ebx
	addl win_y1, %edx
	
	movl %ebx, pos_x
	movl %edx, pos_y
	
	call move_cursor
	
	pop %ebx
	leave
	ret

# Internal, cursor update function
# ebx = x coordinate
# edx = y coordinate
move_cursor:
	movl %edx, %eax
	movl $80, %ecx
	mull %ecx
	addl %ebx, %eax
	
	pushw %ax
	movb $0x0E, %al
	movw $0x3D4, %dx
	outw %ax, %dx
	popw %ax
	xchgb %al, %ah
	movb $0x0F, %al
	outw %ax, %dx

	ret

.macro _set_cursor value1 value2
	movw \value1, %ax
	movw $0x3D4, %dx
	outw %ax, %dx
	movw \value2, %ax
	outw %ax, %dx
.endm

hide_cursor:
	_set_cursor $0x200A $0x000B
	ret

show_cursor:
	_set_cursor $0x0E0A $0x0F0B
	ret

# void highlight(int x1, int x2, int y, int atr)
# highlight selected portion of the line y with color attribute atr
highlight:
	push %ebp
	movl %esp, %ebp
	
	push %es
	push %gs
	pop %es
	push %edi
	movl 16(%ebp), %eax
	movl $80, %ecx
	mull %ecx
	addl 8(%ebp), %eax
	shll %eax
	incl %eax
	movl %eax, %edi		# edi = (y*80 + x1)*2 + 1
	movl 20(%ebp), %eax	# eax = atr
	
	movl 12(%ebp), %ecx
	subl 8(%ebp), %ecx
	incl %ecx		# ecx = x2 - x1 + 1
	
_for1:	
	stosb			# [es:edi] = al edi++
	incl %edi
	loop _for1
	
	pop %edi
	pop %es
	leave
	ret

# The exclusive text-output function
write:
	push %ebp
	movl %esp, %ebp
	
	push %es
	push %gs
	pop %es
	push %ebx
	push %esi
	push %edi
	movl win_x1, %edx
	movl win_x2, %ebx
	
	movl 8(%ebp), %esi
	test %esi, %esi
	jz _end_wr
	
_for3:	movl pos_y, %eax
	movl $80, %ecx
	push %edx
	mull %ecx
	pop %edx
	addl pos_x, %eax
	shll %eax
	movl %eax, %edi
	movb textattr, %ah
	
_for2:	movb (%esi), %al
	incl %esi
	cmpb $0, %al
	je _end_wr
	cmpb $0x0A, %al
	je _nline0
	stosw
	incl pos_x
	cmpl %ebx, pos_x
	jle _for2
	jmp _nline
_nline0:
	cmpl $0, use_eol
	jnz _end_wr
_nline:
	movl %edx, pos_x
	incl pos_y
	jmp _for3
_end_wr:
	movl pos_y, %edx
	movl pos_x, %ebx
	call move_cursor
	
	pop %edi
	pop %esi
	pop %ebx
	pop %es
	leave
	ret

# void put_char(int ch)
# Put one character on the screen and update cursor
put_char:
	push %ebp
	movl %esp, %ebp

	push %es
	push %gs
	pop %es
	push %ebx
	push %edi
	movl win_x1, %edx
	movl win_x2, %ebx

	movl pos_y, %eax
	movl $80, %ecx
	push %edx
	mull %ecx
	pop %edx
	addl pos_x, %eax
	shll %eax
	movl %eax, %edi

	movb textattr, %ah
	movb 8(%ebp), %al
	movw %ax, %es:(%edi)
	incl pos_x
	cmpl %ebx, pos_x
	jle _end_putchar

	movl %edx, pos_x
	incl pos_y

_end_putchar:
	movl pos_y, %edx
	movl pos_x, %ebx
	call move_cursor

	pop %edi
	pop %ebx
	pop %es
	leave
	ret

# void window(x1, y1, x2, y2)
window:
	push %ebp
	movl %esp, %ebp
	
	movl 8(%ebp), %eax
	movl %eax, win_x1
	movl %eax, pos_x
	movl 12(%ebp), %eax
	movl %eax, win_y1
	movl %eax, pos_y
	movl 16(%ebp), %eax
	movl %eax, win_x2
	movl 20(%ebp), %eax
	movl %eax, win_y2
	
	leave
	ret

# void get_window(&x1, &y1, &x2, &y2)
get_window:
	enter $0,$0
	
	movl win_x1, %eax
	movl %eax, 8(%ebp)
	movl win_y1, %eax
	movl %eax, 12(%ebp)
	movl win_x2, %eax
	movl %eax, 16(%ebp)
	movl win_y2, %eax
	movl %eax, 20(%ebp)
	
	leave
	ret

# Fast screen save
# Needed for interrupts which garble the screen
save_screen:
	push %ds
	push %es
	movw %gs, %ax
	movw %ax, %ds
	movw %ax, %es

	movl %esi, %eax
	movl %edi, %edx
	
	xorl %esi, %esi
	movl $0xFA0, %edi
	movl $0x3E8, %ecx
	rep movsl
	
	movl %edx, %edi
	movl %eax, %esi

	pop %es
	pop %ds
	ret

# Fast screen restore
# Needed for interrupts which garble the screen
load_screen:
	push %ds
	push %es
	movw %gs, %ax
	movw %ax, %ds
	movw %ax, %es

	movl %esi, %eax
	movl %edi, %edx
	
	xorl %edi, %edi
	movl $0xFA0, %esi
	movl $0x3E8, %ecx
	rep movsl
	
	movl %edx, %edi
	movl %eax, %esi

	pop %es
	pop %ds
	ret

# Convert unsigned integer to decimal string
inttostr:
	push %ebp
	movl %esp, %ebp
	subl $16, %esp	# Local array
	
	push %ebx
	push %esi
	movl 8(%ebp), %eax
	movl 12(%ebp), %ebx
	movl $10, %esi
	xorl %ecx, %ecx
	decl %ecx

	# Divide number by 10 and store remainders into local array (index is ecx)
_for4:
	xorl %edx, %edx
	divl %esi
	addb $48, %dl
	movb %dl, (%ebp,%ecx,1)
	decl %ecx
	cmpl $0, %eax
	jnz _for4
	
	xorl %edx, %edx
	# Reverse the string from local array into ebx (indices are ecx, edx)
_for5:
	incl %ecx
	movb (%ebp,%ecx,1), %al
	movb %al, (%ebx,%edx,1)
	incl %edx
	cmpl $-1, %ecx
	jnz _for5

	movb $0, (%ebx,%edx,1)
	movl %ebx, %eax

	pop %esi
	pop %ebx
	leave
	ret

# Convert unsigned integer to hexadecimal string
hextostr:
	push %ebp
	movl %esp, %ebp
	subl $16, %esp	# Local array
	
	push %ebx
	push %esi
	movl 8(%ebp), %eax
	movl 12(%ebp), %ebx
	movl $16, %esi
	xorl %ecx, %ecx
	decl %ecx

	# Divide number by 16 and store remainders into local array (index is ecx)
_for41:
	xorl %edx, %edx
	divl %esi
	cmpl $9, %edx
	jle _dec
	addb $7, %dl	# correct for HEXA characters (A-F)
_dec:	addb $48, %dl   # correct for HEXA and DEC characters
	movb %dl, (%ebp,%ecx,1)
	decl %ecx
	cmpl $0, %eax
	jnz _for41
# Attach "0x" prefix before to indicate hexa. (here in reverse order)
	movb $120, (%ebp,%ecx,1)
	decl %ecx
	movb $48, (%ebp,%ecx,1)
	decl %ecx

	xorl %edx, %edx
	# Reverse the string from local array into ebx (indices are ecx, edx)
_for51:
	incl %ecx
	movb (%ebp,%ecx,1), %al
	movb %al, (%ebx,%edx,1)
	incl %edx
	cmpl $-1, %ecx
	jnz _for51

	movb $0, (%ebx,%edx,1)
	movl %ebx, %eax

	pop %esi
	pop %ebx
	leave
	ret

# Convert decimal string to integer (if possible)
# int strtoint(char *src);
strtoint:
	enter $0,$0
	push %ebx
	push %esi

	movl 8(%ebp), %ebx
	movl $10, %esi

	push %ebx
	call strlen
	addl $4, %esp
	cmpl $10, %eax
	ja _err_strtoint
	cmpl $0, %eax
	ja _cont_strtoint

_err_strtoint:
	movl $-1, %eax
	jmp _end_strtoint
	
_cont_strtoint:
	xorl %ecx, %ecx
	xorl %eax, %eax
_start_strtoint:
	movzbl (%ebx,%ecx,1), %edx
	cmpl $0, %edx
	jz _end_strtoint
	cmpl $48, %edx
	jb _err_strtoint
	cmpl $57, %edx
	ja _err_strtoint
	push %edx
	mull %esi
	pop %edx
	subl $48, %edx
	addl %edx, %eax
	incl %ecx
	jmp _start_strtoint

_end_strtoint:
	pop %esi
	pop %ebx
	leave
	ret

# Convert hexadecimal string to integer (if possible)
# int strtohex(char *src);
strtohex:
	enter $0,$0
	push %ebx
	push %esi

	movl 8(%ebp), %ebx
	movl $16, %esi

	push %ebx
	call strlen
	addl $4, %esp
	cmpl $8, %eax
	jbe _cont_strtohex
_err_strtohex:
	movl $-1, %eax
	jmp _end_strtohex
	
_cont_strtohex:
	xorl %ecx, %ecx
	xorl %eax, %eax
_start_strtohex:
	movzbl (%ebx,%ecx,1), %edx
	cmpl $0, %edx
	jz _end_strtohex
	cmpl $48, %edx	# ascii 0
	jb _err_strtohex
	cmpl $57, %edx  # ascii 9
	jbe _num_strtohex
	cmpl $65, %edx  # ascii A
	jb _err_strtohex
	cmpl $70, %edx  # ascii F
	jbe _ltr_strtohex
	cmpl $97, %edx	# ascii a
	jb _err_strtohex
	cmpl $102, %edx	# ascii f
	ja _err_strtohex

	subl $32, %edx	# correction for a-f letters
_ltr_strtohex:
	subl $7, %edx	# correction for A-F letters
_num_strtohex:
	push %edx
	mull %esi
	pop %edx
	subl $48, %edx
	addl %edx, %eax
	incl %ecx
	jmp _start_strtohex

_end_strtohex:
	pop %esi
	pop %ebx
	leave
	ret

#
pad_right:
	push %ebp
	movl %esp, %ebp

	push %ebx
	push %esi
	movl 8(%ebp), %esi
	movl 12(%ebp), %ebx
	push %ebx
	call strlen
	addl $4, %esp
	subl %eax, %esi
	js _end_pad
	movl %ebx, %edx
	addl %esi, %edx
	incl %eax
	
_for6:
	decl %eax
	movb (%ebx,%eax,1), %cl
	movb %cl, (%edx,%eax,1)
	test %eax, %eax
	jnz _for6

	movl 16(%ebp), %eax
_for7:
	decl %esi
	movb %al, (%ebx,%esi,1)
	test %esi, %esi
	jnz _for7

_end_pad:
	movl %ebx, %eax

	pop %esi
	pop %ebx
	leave
	ret

# Wait for keypress and return its scan-code (only press events, without prefixes)
readkey:
	call empty_8042
	xor %eax, %eax
	inb $0x60, %al
	cmpb $0x7F, %al
	ja readkey
	ret

# Wait for enter key to be pressed
read_enter:
	call readkey
	cmpb $28, %al
	jne read_enter
	ret

# Internal function - poll for key press
empty_8042:
	.word	0x00eb,0x00eb
	inb	$0x64, %al	# 8042 status port
	testb	$1, %al		# is output buffer empty?
	jz	empty_8042	# yes - loop
	ret

# read keyboard scan code and translate into ASCII char
read_char:
	push %ebx
	lea scan2ascii, %ebx
_for8:
	call readkey
	xlatb
	cmpb $0, %al
	je _for8

	pop %ebx
	ret

# char *read_string(char *dest, int size);
# read characters into <dest> until enter pressed or <size> characters was read
read_string:
	push %ebp
	movl %esp, %ebp

	push %ebx
	push %esi
	movl 8(%ebp), %ebx
	movl 12(%ebp), %ecx
	xorl %edx, %edx

_start_read_string:
	call read_char
	cmpb $28, %al
	je _end_read_string
	cmpb $1, %al
	je _esc_read_string
	cmpb $14, %al
	jne _cont_read_string
	cmpl $0, %edx
	je _start_read_string
	movl pos_x, %esi
	cmpl $0, %esi
	je _start_read_string
	decl %esi
	movl %esi, pos_x
	push %edx
	push %ecx
	push %ebx
	subl $4, %esp
	movl $32, (%esp)
	call put_char
	addl $4, %esp
	movl %esi, pos_x
	movl %esi, %ebx
	movl pos_y, %edx
	call move_cursor
	pop %ebx
	pop %ecx
	pop %edx
	decl %edx
	jmp _start_read_string
	
_cont_read_string:
	movb %al, (%ebx,%edx,1)

	push %edx
	push %ecx
	push %eax
	call put_char
	addl $4, %esp
	pop %ecx
	pop %edx

	incl %edx
	cmpl %edx, %ecx
	jg _start_read_string
	jmp _end_read_string

_esc_read_string:
	xorl %edx, %edx
_end_read_string:
	movb $0, (%ebx,%edx,1)
	movl %ebx, %eax

	pop %esi
	pop %ebx
	leave
	ret

scan2ascii:
	.ascii "\000\0011234567890-=\016\000QWERTYUIOP[]\034\000AS"
	.ascii "DFGHJKL;'`\000\\ZXCVBNM,./\000*\000 \000\000\000\000\000\000"
	.ascii "\000\000\000\000\000\000\000789-456+1230."
	.fill 44,1,0
