// Prot386 Descriptor manipulation
// (C) 2006-2007 Peter Ambroz
// This code is free software and is distributed under the terms of GNU GPL

#include "prototypes.h"

char *s1 = "GDT Descriptors\nCreate New/Change\nView/Edit Existing\nBack";
char *s2 = "LDT Descriptors\nCreate New/Change\nView/Edit Existing\nSet Active LDT\nBack";

char *sys_types[16] = {"Reserved","Available TSS 286","LDT","Busy TSS 286",
		 "Call Gate 286","Task Gate","Interrupt Gate 286","Trap Gate 286",
		 "Reserved","Available TSS","Reserved","Busy TSS",
		 "Call Gate","Reserved","Interrupt Gate","Trap Gate"};

// Order of sys_types items in menu
int sys_biject[16] = {12, 5, 14, 15, 9, 11, 2, 4, 6, 7, 1, 3, 0, 8, 10, 13};

char *user_types[2] = {"Data segment","Code segment"};
char *user_flags[2][4] = {{"Expand-up, Read-only","Expand-up, Read/Write",
			   "Expand-down, Read-only","Expand-down, Read/Write"},
			  {"Non-Conforming, Exec-only","Non-Conforming, Readable",
			   "Conforming, Exec-only","Conforming, Readable"}};
char *bit_sizes[2] = {" (16-bit)"," (32-bit)"};
char *presence[2] = {"Not Present","Present"};

char *dt0[3] = {"GDT","LDT","IDT"};
char *dd0 = " Descriptor no. ";
char *suffix[4] = {""," KB)"," MB)"," GB)"};
char *nochange = "Warning\nThis is kernel descriptor.\nYou are not allowed to change it.";
char *nomem = "Warning\nKernel memory (1-st MB)\ncan't be used in descriptors.\nChanges will be dropped.";

char *dlg0 = "\nEnter new value\n";
char *dlg1 = "\nEnter new value\n(hexadecimal)\n";
char *dlg2 = "\nEnter new index\n";
char *dlg3 = "\nEnter GDT index\n";

// idt.s
extern char *ex_names[];

char mnu0[224] = "\0";
char mnu1[16] = "\nGDT\nLDT";
char mnu2[192] = "\0";
char mnu3[240] = "\0";
char info[256], temp[20];

#define LDT_MAX 15
int g2l_biject[LDT_MAX+1];

extern int old1, old4;
int old2, old3;

// Soft protection from several damage to the computer
// Forbids changing of important descriptors and using a memory occupied by kernel
int gdt_protect = 6;	// Last GDT entry, that will be protected against change
unsigned int mem_protect = 0xFFFFF;	// Last physical memory address that can't be used

DESCRIPTOR dtmp, dtmp2;
MENU_STATE mtmp;

/* Read descriptor information */
DESCRIPTOR *load_descriptor(int selector, TABLE tab, DESCRIPTOR *d1) {
  unsigned int desc[2];

  xdt_read(selector, tab, desc);
  d1->index=selector;
  d1->dti=tab;
  d1->dts = dt0[tab];
  d1->sel_min = selector * 8 + d1->dti * 4;
  d1->sel_max = d1->sel_min + 3;

  if (desc[0]==0 && desc[1]==0) {
    d1->s = 0;
    d1->type = 0;
    return d1;
  }

  d1->p = (desc[1] >> 15) & 0x01;
  d1->dpl = (desc[1] >> 13) & 0x03;
  d1->s = (desc[1] >> 12) & 0x01;
  d1->type = (desc[1] >> 8) & 0x0F;

  d1->base = (desc[1] & 0xFF000000) + ((desc[1] & 0xFF) << 16) + (desc[0] >> 16);
  d1->limit = (desc[1] & 0x000F0000) + (desc[0] & 0xFFFF);
  d1->g = (desc[1] >> 23) & 0x01;
  if (d1->g) d1->limit = (d1->limit << 12) + 4095;

  d1->seg = desc[0] >> 16;
  d1->ofs = (desc[1] & 0xFFFF0000) + (desc[0] & 0xFFFF);
  d1->pc = desc[1] & 0x1F;

  if (d1->s) {  // Code or Data segment
    d1->d = (desc[1] >> 22) & 0x01;
  }
  return d1;
}

#define PRINT_BASE_LIMIT strcat(info, "\nBase: ", hextostr(d1->base, tmps), 0); \
			 strcat(info, "\nLimit: ", hextostr(d1->limit, tmps), 0); \
			 if (suf) strcat(info, " (", inttostr(lim, tmps), suffix[suf], 0);
//			 if (d1->g) strcat(info, " (4kB granular)", 0);

#define PRINT_SELECTORS strcat(info, "\nSelectors: ", inttostr(d1->sel_min, tmps), 0); \
			strcat(info, " - ", inttostr(d1->sel_max, tmps), 0);

#define PRINT_SELECTOR strcat(info, "\nSegment: ", dt0[(d1->seg>>2)&1], "[", \
			      inttostr(d1->seg>>3, tmps), "] (", 0); \
		       strcat(info, inttostr(d1->seg, tmps), ")", 0);
#define PRINT_OFFSET strcat(info, "\nOffset: ", hextostr(d1->ofs, tmps), 0);
#define PRINT_PARAMS strcat(info, "\nParameters: ", inttostr(d1->pc, tmps), 0);
#define NEW_LINE strcat(info, "\n", 0);

/* Parse descriptor information */
char *parse_descriptor(DESCRIPTOR *d1, char *info) {
  char tmps[11];
  unsigned int lim, suf=0;

  info[0]='\0';

  strcat(info, d1->dts, dd0, inttostr(d1->index,tmps), 0);
  if ((!d1->s) && (!d1->type)) {
    strcat(info, "\nType: Invalid Descriptor", 0);
    PRINT_SELECTORS
    return info;
  }

  if ((d1->s) && (((d1->type >> 2)&0x03) == 1)) { // Expand-down data segment
    lim = 0xFFFFFFFFu - d1->limit;
  } else lim = d1->limit+1;

  if (lim == 0) { lim = 4; suf = 3; }
  if (lim >= 0x40000000) { lim >>= 30; suf=3; }
  if (lim >= 0x100000) { lim >>= 20; suf=2; }
  if (lim >= 0x400) { lim >>= 10; suf=1; }

  strcat(info, "\nType: ", 0);
  if (d1->s) {  // Code or Data segment
    strcat(info, user_types[d1->type<8?0:1], bit_sizes[d1->d], 0);
    strcat(info, "\nFlags: ", user_flags[d1->type<8?0:1][(d1->type>>1)&0x03], 0);
    PRINT_BASE_LIMIT
  } else {
    strcat(info, sys_types[d1->type], 0);
    switch(d1->type) {
      case 1:  // Available TSS 286
      case 2:  // LDT
      case 3:  // Busy TSS 286
      case 9:  // Available TSS
      case 11: NEW_LINE PRINT_BASE_LIMIT break; // Busy TSS
      case 5: NEW_LINE PRINT_SELECTOR NEW_LINE break; // Task Gate
      case 4:  // Call Gate 286
      case 12: PRINT_PARAMS PRINT_SELECTOR PRINT_OFFSET break; // Call Gate
      case 6:  // Int Gate 286
      case 7:  // Trap Gate 286
      case 14: // Int Gate
      case 15: NEW_LINE PRINT_SELECTOR PRINT_OFFSET break; // Trap Gate
    }
  }
  strcat(info, "\nDPL: ", inttostr(d1->dpl, tmps), 0);
  strcat(info, "\n", presence[d1->p], 0);
  PRINT_SELECTORS

  return info;
}


#define SET_BASE_LIMIT desc[1] |= (d1->base & 0xFF000000) + ((d1->base & 0xFF0000) >> 16); \
		       desc[0] |= (d1->base & 0xFFFF) << 16; \
		       desc[1] |= (limit & 0x000F0000) + ((d1->g & 0x01) << 23); \
		       desc[0] |= (limit & 0xFFFF);

#define SET_SELECTOR desc[0] |= (d1->seg << 16);
#define SET_OFFSET desc[1] |= (d1->ofs & 0xFFFF0000); \
		   desc[0] |= (d1->ofs & 0xFFFF);

#define SET_PARAMS desc[1] |= (d1->pc & 0x1F);

void store_descriptor(DESCRIPTOR *d1) {
  unsigned int desc[2] = {0,0};
  unsigned int limit;

  if ((!d1->s) && (!d1->type)) {  // Invalid descriptor
    xdt_write(d1->index, d1->dti, desc);
    return;
  }

  desc[1] |= (d1->p & 0x01) << 15;
  desc[1] |= (d1->dpl & 0x03) << 13;
  desc[1] |= (d1->s & 0x01) << 12;
  desc[1] |= (d1->type & 0x0F) << 8;

  limit = d1->limit;
  if (d1->g) limit = d1->limit >> 12;
  else if (d1->limit > 0xFFFFF) {
    limit = d1->limit >> 12;
    d1->g = 1;
  }

  // Kernel memory protection:
  if (((!d1->s) && (((d1->type >> 2)&0x01) == 0)) || (d1->s))
  { // Only those descriptors with BASE and LIMIT fields (system with bit 2 in type flags clear, or non-system)
    if (d1->base <= mem_protect) { // BASE is inside kernel protected memory
      okno(18, 50, 6, 4, 116, 0, nomem);
      read_enter();
      return;
    }
    if ((d1->s) && (((d1->type >> 2)&0x03) == 1))
    { // Expand-down data segments need to consider BASE and LIMIT differently
      if (mem_protect - d1->base > d1->limit)
      { // LIMIT falls into kernel memory
	okno(18, 50, 6, 4, 116, 0, nomem);
	read_enter();
	return;
      }
    } else {  // Expand-up segments
      if (0xFFFFFFFFu - d1->base < d1->limit)
      { // LIMIT falls into kernel memory
	okno(18, 50, 6, 4, 116, 0, nomem);
	read_enter();
	return;
      }
    }
  }

  if (d1->s) {
    desc[1] |= (d1->d & 0x01) << 22;
    SET_BASE_LIMIT
  } else {
    switch(d1->type) {
      case 1:  // Available TSS 286
      case 2:  // LDT
      case 3:  // Busy TSS 286
      case 9:  // Available TSS
      case 11: SET_BASE_LIMIT break; // Busy TSS
      case 5: SET_SELECTOR break; // Task Gate
      case 4:  // Call Gate 286
      case 12: SET_SELECTOR SET_OFFSET SET_PARAMS break; // Call Gate
      case 6:  // Int Gate 286
      case 7:  // Trap Gate 286
      case 14: // Int Gate
      case 15: SET_SELECTOR SET_OFFSET break; // Trap Gate
    }
  }

  xdt_write(d1->index, d1->dti, desc);
}

void write_info(char *info) {
  poz(0,0);
  okno(21, 58, 8, 7, 112, 0, info);
}

char *summary(int sel, char *data);
char *lsummary(int sel, char *data);

void nop(DESCRIPTOR *d1) {}

void ed_flags(DESCRIPTOR *d1) {
  int typ, index;
  typ = d1->type<8?0:1;
  index = (d1->type >> 1) + 2 + typ;
  index = menu(10, 40, 6, 10, 48, index, 0, AL_LEFT, mnu0);
  if (index == -1) return;
  if (index > 5) { typ = 8; index -= 5; }
  else typ = 0;
  if (index == 1) index++;
  index-=2;
  d1->type = typ + (index << 1);
}

void ed_tseg(DESCRIPTOR *d1) {
  int val;
  val = menu_ex(32, 69, 7, 8, 48, &mtmp, 256, dt0[0], summary, 0);
  if (val == -1) return;
  d1->seg = val << 3;
}

void ed_seg(DESCRIPTOR *d1) {
  int val, dt;
  dt = ((d1->seg >> 2) & 0x01) + 1;
  dt = menu(22, 31, 7, 2, 48, dt, 0, AL_LEFT, mnu1);
  switch (dt) {
    case -1: val = -1; break;
    case 1: val = menu_ex(32, 69, 7, 8, 48, &mtmp, 256, dt0[0], summary, 0); break;
    case 2: val = menu_ex(32, 69, 7, 8, 48, &mtmp, 256, dt0[1], lsummary, 0); break;
  }
  if (val == -1) return;
  d1->seg = (val << 3) + ((dt-1) << 2);
}

void ed_ofs(DESCRIPTOR *d1) {
  unsigned int val;
  val = HEXDIALOG(3, dlg1);
  d1->ofs = val;
}

void ed_base(DESCRIPTOR *d1) {
  unsigned int val;
  val = HEXDIALOG(3, dlg1);
  d1->base = val;
}

void ed_limit(DESCRIPTOR *d1) {
  unsigned int val;
  val = HEXDIALOG(3, dlg1);
  d1->limit = val;
  if (val > 0xFFFFF) {
    d1->g = 1;
    d1->limit = ((val >> 12) << 12) + 4095;
  } else d1->g = 0;
}

void ed_pc(DESCRIPTOR *d1) {
  int val;
  val = DIALOG(2, dlg0);
  if ((val < 0) || (val > 31)) return;
  d1->pc = val;
}

void ed_dpl(DESCRIPTOR *d1) {
  int val;
  val = DIALOG(2, dlg0);
  if ((val < 0) || (val > 3)) return;
  d1->dpl = val;
}

void ed_p(DESCRIPTOR *d1) {
  d1->p ^= 1;
}

void edit_descriptor(DESCRIPTOR *d1) {
  void (*fx[8])(DESCRIPTOR *) = {nop,nop,nop, ed_base, ed_limit, ed_dpl, ed_p, nop};

  if ((d1->dti==TB_GDT) && (d1->index <= gdt_protect))
  { // SW protection of essential descriptors.
    okno(18, 54, 6, 2, 116, 0, nochange);
    read_enter();
    return;
  }

  if (d1->s) {
    fx[2] = ed_flags;
  } else {
    switch(d1->type) {
      case 5: fx[3] = ed_tseg; fx[4] = nop; break; // Task Gate
      case 4:  // Call Gate 286
      case 12: fx[2] = ed_pc; fx[3] = ed_seg; fx[4] = ed_ofs; break; // Call Gate
      case 6:  // Int Gate 286
      case 7:  // Trap Gate 286
      case 14: // Int Gate
      case 15: fx[3] = ed_seg; fx[4] = ed_ofs; break; // Trap Gate
    }
  }

  do {
    poz(0,0);
    old2=menu(21,58,8,7,112,old2,0,AL_LEFT,info);
    if (old2 == -1) break;
    fx[old2](d1);
    parse_descriptor(d1, info);
  } while(1);
  store_descriptor(d1);
}

void change_type(DESCRIPTOR *d1) {
  if ((d1->dti==TB_GDT) && (d1->index <= gdt_protect))
  { // SW protection of essential descriptors.
    okno(18, 54, 6, 2, 116, 0, nochange);
    read_enter();
    return;
  }
  old3=menu(17,39,4,15,48,1,0,AL_LEFT,mnu2);
  switch(old3) {
    case -1: return;
    case 0: break;
    case 1: d1->s = 0; d1->type = 0; break;
    case 2: d1->s = 1; d1->type = 2; break;
    case 3: d1->s = 1; d1->type = 10; break;
    default: d1->s = 0; d1->type = sys_biject[old3-4]; break;
  }
  d1->d = 1; d1->dpl = 3; d1->p = 1;
  d1->base = 0x100000; d1->limit = 0x1FFFFF; d1->g = 1;
  d1->seg = 56; d1->ofs = 0; d1->pc = 0;
  if (sys_biject[old3-4] == 5) d1->seg = 32; // Task gate
  if (sys_biject[old3-4] == 2) {	// LDT - allocate extra space for each LDT
    d1->base = 0x300000 + (d1->index-4)*4096; d1->limit = 0xFFF; d1->g = 0;
  }
  store_descriptor(d1);
}

char *summary(int sel, char *data) {
  sel%=256;
  inttostr(sel, data);
  strcat(data, ": ", 0);
  load_descriptor(sel, TB_GDT, &dtmp2);

  if ((!dtmp2.s) && (!dtmp2.type)) {
    strcat(data, "Invalid Descriptor (Empty)", 0);
    return data;
  }
  if (dtmp2.s) strcat(data, user_types[dtmp2.type<8?0:1], bit_sizes[dtmp2.d], 0);
  else strcat(data, sys_types[dtmp2.type], 0);

  return data;
}

char *lsummary(int sel, char *data) {
  sel%=256;
  inttostr(sel, data);
  strcat(data, ": ", 0);
  load_descriptor(sel, TB_LDT, &dtmp2);

  if ((!dtmp2.s) && (!dtmp2.type)) {
    strcat(data, "Invalid Descriptor (Empty)", 0);
    return data;
  }
  if (dtmp2.s) strcat(data, user_types[dtmp2.type<8?0:1], bit_sizes[dtmp2.d], 0);
  else strcat(data, sys_types[dtmp2.type], 0);

  return data;
}

char *isummary(int sel, char *data) {
  sel%=32;
  hextostr(sel, data);
  if (sel > 19) sel = 15;
  strcat(data, ": ", 0);
  strcat(data, ex_names[sel], 0);

  return data;
}

void init_menus(void) {
  int i,j;

  if (!mnu0[0]) {   // Initialize flags menu
    strcat(mnu0, "\n", 0);
    for (i=0; i<2; i++) {
      strcat(mnu0, user_types[i], "\n", 0);
      for (j=0; j<4; j++) strcat(mnu0, " ", user_flags[i][j], "\n", 0);
    }
    mnu0[strlen(mnu0)-1] = '\0';
  }

  if (!mnu2[0]) {
    strcat(mnu2, "\nEmpty\n", 0);
    for (i=0; i<2; i++) strcat(mnu2, user_types[i], "\n", 0);
    for (i=0; i<12; i++) strcat(mnu2, sys_types[sys_biject[i]], "\n", 0);
    mnu2[strlen(mnu2)-1] = '\0';
  }
}

void descriptors(void)
{
  int ch;
  static int sel = 0;

  init_menus();
  mtmp.ht = 1; mtmp.it = 0;

  do {
    switch(old1=menu(36,58,5,0,112,old1,0,AL_LEFT,s1)) {
      case 1:
	save_screen();
	do {
	  poz(0,0);
	  old2 = menu_ex(21, 58, 7, 8, 112, &mtmp, 256, dt0[0], summary, 0);
	  if (old2 == -1) break;
	  load_descriptor(old2, TB_GDT, &dtmp);
	  change_type(&dtmp);
        } while (1);
	load_screen();
	break;
      case 2:
	mtmp.ht = 1; mtmp.it = 0;
	save_screen();
	poz(0,0);
	do {
	  load_descriptor(sel, TB_GDT, &dtmp);
	  write_info(parse_descriptor(&dtmp, info));
	  ch = readkey();
	  switch(ch) {
	    case 80: sel--; if (sel < 0) sel=255; break;
	    case 72: sel++; if (sel > 255) sel=0; break;
	    case 28: old2=1; edit_descriptor(&dtmp); break;
	  }
	} while (ch != 1);
	load_screen();
        break;
      case 3: return;
      case -1: return;
    }
  } while (1);
  return;
}

int list_ldt(int active) {
  int i, cnt=0;
  char tmps[11];

  mnu3[0] = '\0';
  load_descriptor(active, TB_GDT, &dtmp);
  if ((dtmp.s == 0) && (dtmp.type == 2)) {
    strcat(mnu3, "Active: ", inttostr(active, tmps), 0);
    strcat(mnu3, ", Base=", hextostr(dtmp.base, tmps), "\n", 0);
  } else strcat(mnu3, "Last Active: ", inttostr(active, tmps), ", REMOVED\n", 0);
  for (i=1; i<256; i++) {
    load_descriptor(i, TB_GDT, &dtmp);
    if ((dtmp.s == 0) && (dtmp.type == 2)) {
      strcat(mnu3, inttostr(i, tmps), ": ", 0);
      strcat(mnu3, "Base=", hextostr(dtmp.base, tmps), "\n", 0);
      cnt++;
      g2l_biject[cnt] = i;
    }
    if (cnt == 15) break;
  }
  if (!cnt) return -1;
  mnu3[strlen(mnu3)-1] = '\0';
  return 0;
}

void ldescriptors(void)
{
  int ch;
  static int sel = 0;

  init_menus();
  mtmp.ht = 1; mtmp.it = 0;

  do {
    switch(old4=menu(36,58,5,0,112,old4,0,AL_LEFT,s2)) {
      case 1:
	save_screen();
	do {
	  poz(0,0);
	  old2 = menu_ex(21, 58, 7, 8, 112, &mtmp, 256, dt0[1], lsummary, 0);
	  if (old2 == -1) break;
	  load_descriptor(old2, TB_LDT, &dtmp);
	  change_type(&dtmp);
        } while (1);
	load_screen();
	break;
      case 2: 
	save_screen();
	poz(0,0);
	do {
	  load_descriptor(sel, TB_LDT, &dtmp);
	  write_info(parse_descriptor(&dtmp, info));
	  ch = readkey();
	  switch(ch) {
	    case 80: sel--; if (sel < 0) sel=255; break;
	    case 72: sel++; if (sel > 255) sel=0; break;
	    case 28: old2=1; edit_descriptor(&dtmp); break;
	  }
	} while (ch != 1);
	load_screen();
	break;
      case 3:
	save_screen();
	poz(0,0);
	if (list_ldt(load_ldtr())) {	// No LDT descriptors in GDT
	  okno(21, 48, 7, 4, 116, "Warning", "No LDT descriptor found.\nCreate at least 1 LDT\ndescriptor in GDT");
	  read_enter();
	  load_screen();
	  return;
	}
	old2 = menu(21, 54, 7, 0, 112, 1, 0, AL_LEFT, mnu3);
	if (old2 > 0) store_ldtr(g2l_biject[old2]);
	load_screen();
	break;
      case 4: return;
      case -1: return;
    }
  } while (1);
  return;
}

void idescriptors(void)
{
  save_screen();
  mtmp.ht = 1; mtmp.it = 0;
  do {
    poz(0,0);
    old2 = menu_ex(21, 58, 7, 8, 112, &mtmp, 32, dt0[2], isummary, 0);
    if (old2 == -1) break;
    sw_int(old2);
  } while (1);
  load_screen();
  return;
}
